import platform
from typing import Union
from ..core import Note, Clip, Track, Midi
from ..core import item2midi

from mido import Backend, get_output_names

__all__ = ["Controller"]


class Controller:

    def __init__(self, device_name, backend_name='default'):

        if backend_name == 'default':
            platform_name = platform.system()
            if (platform_name == 'Windows'):
                backend_name = 'mido.backends.rtmidi/WINDOWS_MM'
            elif platform_name == 'Linux':
                backend_name = 'mido.backends.rtmidi/LINUX_ALSA'
            elif platform_name == 'Darwin':
                backend_name = 'mido.backends.rtmidi/MACOSX_CORE'

        self.name = device_name
        self.backend = Backend(f'{backend_name}')
        self.port = self.backend.open_output(device_name)

    @classmethod
    def get_avaliable_devices(cls):
        return get_output_names()

    def send(self, msg):
        self.port.send(msg)

    def play(self, mid: Union[Note, Clip, Track, Midi]):
        for msg in mid.to_mido().play():
            self.send(msg)

    def __call__(self, item: Union[Note, Clip, Track, Midi], **kwargs) -> None:
        self.play(item2midi(item, **kwargs))



import math
import re

from typing import Iterable, List, Optional, Union
from mido import Message

from ..exceptions import ParameterError
from . import event
from .clip import Clip
from .midi import Midi
from .event import Event
from .note import Note
from .track import Track
from ..constants import DRUM_SOUNDS, KEY_TO_INDEX, KEYS, DEFAULT_TICKS_PER_BEAT, MESSAGE_TYPE_TO_EVENT_TYPE

__all__ = [
    "loc2tick",
    "tick2loc",
    "loc2second",
    "second2loc",
    "value2name",
    "name2value",
    "value2frequency",
    "frequency2value",
    "key2index",
    "message2event",
    "messages2events",
    "note2clip",
    "note2track",
    "note2midi",
    "clip2track",
    "clip2track",
    "clip2midi",
    "track2midi",
    "item2midi",
]


def loc2tick(loc: float, ticks_per_beat: int) -> int:
    return round(loc * ticks_per_beat)


def tick2loc(ticks, ticks_per_beat: int) -> float:
    return ticks / ticks_per_beat


def loc2second(loc: float, bpm: float) -> float:
    return 60 / bpm * loc


def second2loc(second: float, bpm: float) -> float:
    return second * bpm / 60


# borrowed from librosa
def name2value(name: str) -> int:
    match = re.match(
        r'^(?P<key>[A-Ga-g])'
        r'(?P<acc>[#b]*)'
        r'(?P<oct>[+-]?\d+)?', name)
    if not match:
        raise ParameterError(f'Invalid note name: {name}')

    key = match.group('key').upper()
    acc = match.group('acc')
    oct = match.group('oct')

    # default to 4.
    oct = int(oct) if oct else 4

    value = KEY_TO_INDEX[key] + 12 * (oct + 1)

    for ch in acc:
        if ch == '#':
            value += 1
        elif ch == 'b':
            value -= 1

    return value


def value2name(value: int) -> str:
    key = KEYS[value % 12]
    oct = int(value / 12) - 1
    name = f'{key:s}{oct:0d}'
    return name


def value2frequency(value: int) -> float:
    return 440.0 * (2.0**((value - 69.0) / 12.0))


def frequency2value(freq: float) -> int:
    return round(12 * (math.log2(freq) - math.log2(440.0)) + 69)


def key2index(key:str, root:str='C') -> int:
    key_idx = KEYS.index(key)
    root_idx = KEYS.index(root)
    return (key_idx - root_idx) % 12 


def drum2value(drum: str) -> int:
    try:
        idx = DRUM_SOUNDS.index(drum)
    except:
        raise ParameterError(f"Invalid drum name: {drum}.")

    return idx + 35


def message2event(
        msg: Message,
        ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> Optional[Event]:
    if msg.type in MESSAGE_TYPE_TO_EVENT_TYPE:
        return eval("event." + MESSAGE_TYPE_TO_EVENT_TYPE[msg.type]).from_message(
            msg, ticks_per_beat)
    else:
        return None


def messages2events(
        msgs: Iterable[Message],
        ticks_per_beat: int = DEFAULT_TICKS_PER_BEAT) -> List[Event]:
    events = []
    cur_time = 0
    for msg in msgs:
        cur_time += msg.time
        msg.time = cur_time
        event = message2event(msg, ticks_per_beat)
        events.append(event)
    return events


def note2clip(note: Note, **kwargs) -> Clip:
    return Clip([note], **kwargs)


def note2track(note: Note, **kwargs) -> Track:
    return clip2track(note2clip(note, **kwargs), **kwargs)


def note2midi(note: Note, **kwargs) -> Midi:
    return track2midi(note2track(note, **kwargs), **kwargs)


def clip2track(clip: Clip, **kwargs) -> Track:
    track = Track(**kwargs)
    track.add(clip)
    return track


def clip2midi(clip: Clip, **kwargs) -> Midi:
    return track2midi(clip2track(clip, **kwargs), **kwargs)


def track2midi(track: Track, **kwargs) -> Midi:
    return Midi([track], **kwargs)


def item2midi(item: Union[Note, Clip, Track, Midi], **kwargs) -> Midi:
    if isinstance(item, Midi):
        mid = item
    elif isinstance(item, Track):
        mid = track2midi(item, **kwargs)
    elif isinstance(item, Clip):
        mid = clip2midi(item, **kwargs)
    elif isinstance(item, Note):
        mid = note2midi(item, **kwargs)
    else:
        raise ParameterError(f"Invalid item type:{type(item)}.")
    return mid



import re
from typing import Callable, Dict

from .. import Clip, Note
from ..core.convert import drum2value
from ._chord import chord

__all__ = [
    'sheet',
]


def _chord_and_note_sheet(item: str, length: float):

    def _note(note, length):
        return Note(note, length=length)

    def _chord(chord_name, length):
        root, chord_type = chord_name.split(':')
        return chord(root, chord_type, length=length)

    if item.isdigit():
        return _note(int(item), length)
    if re.match('[ABCDEFG][b#]*[0-9]$', item):
        return _note(item, length)
    if re.match('[ABCDEFG][b#]*[0-9]:\S*$', item):
        return _chord(item, length)


DEFAULT_MAPPING = {
    # Kick
    'K':
    lambda _, length: Note(
        drum2value('Acoustic Bass Drum'), length=length, channel=9),
    # Flore Tom
    'FT':
    lambda _, length: Note(
        drum2value('Low Floor Tom'), length=length, channel=9),
    # Tom Drum1
    'T':
    lambda _, length: Note(drum2value('High Tom'), length=length, channel=9),
    'HT':
    lambda _, length: Note(drum2value('High Tom'), length=length, channel=9),
    # Tom Drum2
    'LT':
    lambda _, length: Note(drum2value('Low Tom'), length=length, channel=9),
    # Snare Drum
    'S':
    lambda _, length: Note(
        drum2value('Acoustic Snare'), length=length, channel=9),
    # Ride Cymbal
    'R':
    lambda _, length: Note(
        drum2value('Ride Cymbal 1'), length=length, channel=9),
    # Closed Hi-Hat
    'H':
    lambda _, length: Note(
        drum2value('Closed Hi Hat'), length=length, channel=9),
    'CH':
    lambda _, length: Note(
        drum2value('Closed Hi Hat'), length=length, channel=9),
    # Open Hi-Hat
    'OH':
    lambda _, length: Note(drum2value('Open Hi Hat'), length=length, channel=9
                           ),
    # Hi-Hat Pedal
    'PH':
    lambda _, length: Note(
        drum2value('Pedal Hi Hat'), length=length, channel=9),
    # Crash Cymbal
    'CC':
    lambda _, length: Note(
        drum2value('Crash Cymbal 1'), length=length, channel=9),
    # note and chord
    '*':
    _chord_and_note_sheet
}


def sheet(text: str,
          mapping: Dict = DEFAULT_MAPPING,
          length_per_bar: float = 4.,
          bar_symbol: str = '|',
          rest_symbol: str = '0',
          delay_symbol: str = '-') -> Clip:
    res, n_bars = _parse_sheet(text, bar_symbol, delay_symbol)
    clip = Clip()
    loc = 0
    new_res = []

    for key, ration in res:
        length = ration * length_per_bar
        if key == rest_symbol:
            pass
        elif key == delay_symbol:
            if new_res:
                new_res[-1][1] += length
        else:
            new_res.append([key, length, loc])

        loc += length

    for key, length, loc in new_res:

        if key in mapping:
            val = mapping[key]
        else:
            val = mapping['*']

        if isinstance(val, Callable):
            val = val(key, length)
            clip.add(val, loc)
        elif isinstance(val, Note):
            clip.add(val, loc)
        elif isinstance(val, Clip):
            clip.add(val, loc)
    return clip


def _parse_sheet(sheet, bar_symbol, delay_symbol):
    bars = sheet.split(bar_symbol)
    n_bars = len(bars)
    res = []
    for bar in bars:
        bar = bar.strip()
        res.extend(_parse_bar(bar.strip(), delay_symbol))
    return res, n_bars


def _parse_bar(bar: str, delay_symbol):
    bar = _read_from_tokens(_tokenize(bar))

    def _parse(bar, ration):
        res = []
        l0 = 1 / len(bar) * ration
        for item in bar:
            if isinstance(item, list):
                res.extend(_parse(item, l0))
            elif item == delay_symbol:
                if res:
                    res[-1][1] += l0
                else:
                    res.append([delay_symbol, l0])
            else:
                res.append([item, l0])
        return res

    return _parse(bar, 1)


# borrowed from https://github.com/fluentpython/lispy
def _tokenize(s):
    s = f'({s})'
    s = s.replace('(', ' ( ').replace(')', ' ) ')
    s = s.split()
    return s


def _read_from_tokens(tokens):
    token = tokens.pop(0)
    if token == '(':
        exp = []
        while tokens and tokens[0] != ')':
            exp.append(_read_from_tokens(tokens))
        tokens.pop(0)
        return exp
    else:
        return token



from typing import Tuple, Union

import numpy as np
from matplotlib import pyplot as plt

from .core import Clip, Midi, Note, Track, item2midi

__all__ = ["plot"]

PLOT_DEFAULT_MIN_LENGTH = 4

def get_piano_roll(clip: Clip, length: int, resolution: int=64):
    # length = max(clip.length, PLOT_DEFAULT_MIN_LENGTH)
    piano_roll = np.zeros((128, int(resolution * length)))
    for note in clip.notes:
        start = int(note.location*resolution)
        end = int((note.location + note.length)*resolution)
        piano_roll[note.value, start:end-1] = note.velocity
    return piano_roll
    
def plot_midi(
    mid: Midi,
    grid_axis: str = "both",
    grid_linestyle: str = ":",
    grid_linewidth: float = 0.5,
    resolution: int = 24,
    figsize: Tuple[float] = (10, 6.18),
    **kwargs,
):

    if not mid.tracks:
        raise RuntimeError("There is no track to plot.")

    cmaps = ("Blues", "Oranges", "Greens", "Reds", "Purples", "Greys")
    n_tracks = len(mid.tracks)
    length = max([track.notes_length for track in mid.tracks] + [PLOT_DEFAULT_MIN_LENGTH])
    xticklabes = np.arange(0, length, 4**(int(length**0.25) - 1))

    if n_tracks > 1:
        fig, axs_ = plt.subplots(n_tracks, figsize=(figsize[0], figsize[1] * n_tracks), sharex=False)
        fig.subplots_adjust(hspace=0)
        axs = axs_.tolist()
    else:
        fig, ax = plt.subplots(figsize=(10, 6.18))
        axs = [ax]


    # Plot tracks
    for idx, (track, ax) in enumerate(zip(mid.tracks, axs)):
        
            piano_roll = get_piano_roll(track.clip, length, resolution)
    
            ax.imshow(
                piano_roll,
                cmap=cmaps[idx % len(cmaps)],
                aspect="auto",
                vmin=0,
                vmax=127,
                origin="lower",
                interpolation="none",
                **kwargs,
            )
            ax.tick_params(
                direction='in',
                bottom=False,
                top=False,
                left=True,
                right=False,
            )

            # Format x-axis
            ax.set_xticks(resolution * xticklabes - 0.5)
            ax.set_xticklabels("")
            # ax.tick_params(axis="x", which="major", width=0)

            # Format y-axis
            ax.set_yticks(np.arange(0, 128, 12))
            ax.set_yticklabels(["C{}".format(i - 1) for i in range(11)])

            # Format axis labels
            ax.set_ylabel(f"{track.name}\n\nnote")

            # Plot the grid
            ax.grid(
                axis=grid_axis,
                color="k",
                linestyle=grid_linestyle,
                linewidth=grid_linewidth,
            )
    else:
        ax.set_xticks(resolution * xticklabes - 0.5)
        ax.set_xticklabels(xticklabes)
        ax.set_xlabel("location (beat)")


def plot(item: Union[Note, Clip, Track, Midi], **kwargs) -> None:
    plot_midi(item2midi(item.copy(), **kwargs), **kwargs)


import os
import pkg_resources
from tempfile import TemporaryDirectory
from typing import Optional, Union
from ..core import Note, Clip, Track, Midi
from ..core import item2midi

__all__ = ["FluidSynth"]


class FluidSynth:

    def __init__(self,
                 font_path: Optional[str] = None,
                 audio_driver: Optional[str] = None):
        
        if not font_path:
            font_path = pkg_resources.resource_filename(
                __name__, 'TimGM6mb.sf2')

        self.font_path = font_path
        self.audio_driver = audio_driver

    def play(self, mid: Midi) -> None:
        with TemporaryDirectory(prefix='.infmidi.FluidSynth.') as dirname:
            midi_path = os.path.join(dirname, 'tmp.mid')
            mid.save(midi_path)
            cmd = 'fluidsynth'
            if self.audio_driver:
                cmd += f'-a {self.audio_driver}'
            cmd += f' -i {self.font_path} {midi_path}'
            os.system(cmd)

    def __call__(self, item: Union[Note, Clip, Track, Midi], **kwargs) -> None:
        self.play(item2midi(item, **kwargs))



from .event import *
from .note import *
from .clip import *
from .track import *
from .midi import *
from .convert import *

__all__ = []
__all__.extend(event.__all__)
__all__.extend(note.__all__)
__all__.extend(clip.__all__)
__all__.extend(track.__all__)
__all__.extend(midi.__all__)
__all__.extend(convert.__all__)




from functools import wraps
import inspect

# ref:https://python3-cookbook.readthedocs.io/zh_CN/latest/c09/p11_write_decorators_that_add_arguments_to_functions.html
class optional_inplace:
    def __init__(self, default: bool=True):
        self.default = default

    def __call__(self, func):

        if 'inplace' in inspect.getfullargspec(func).args:
            raise TypeError('inplace argument already defined')

        @wraps(func)
        def wrapper(obj, *args, inplace: bool = self.default, **kwargs):
            if not inplace:
                obj = obj.copy()
            return func(obj, *args, **kwargs)

        sig = inspect.signature(func)
        parms = list(sig.parameters.values())
        parms.append(
            inspect.Parameter('inplace',
                            inspect.Parameter.KEYWORD_ONLY,
                            default=True))
        wrapper.__signature__ = sig.replace(parameters=parms)
        return wrapper


from itertools import chain

from setuptools import find_packages, setup

import infmidi

EXTRAS_REQUIRE = {
    'plot': ["numpy>=1.12.0", "matplotlib>=1.5"],
    'devices': ["python-rtmidi>=1.1.0"]
}
EXTRAS_REQUIRE['all'] = list(set(chain(*EXTRAS_REQUIRE.values())))

with open('README.rst') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name=infmidi.__title__,
    version=infmidi.__version__,
    author="gongyibei",
    author_email="gongyibei@gmail.com",
    license="MIT",
    description=
    "Manipulate midi and create music with a simple and high level interface.",
    long_description=LONG_DESCRIPTION,
    url="https://github.com/gongyibei/infmidi",
    project_urls={
        "Documentation": "https://infmidi.readthedocs.io/en/latest/",
        "Code": "https://github.com/gongyibei/infmidi",
        "Issue tracker": "https://github.com/gongyibei/infmidi/issues",
    },
    packages=find_packages(),
    package_data={'': ['TimGM6mb.sf2']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Multimedia :: Sound/Audio :: MIDI",
    ],
    install_requires=[
        "mido>=1.1.16",
        "sortedcontainers",
    ],
    extras_require=EXTRAS_REQUIRE,
)
